from natch.abstract import Hasher


class QualnameHasher(Hasher):

    def hash(self, obj):
        hash = '.'.join(
            [
                obj.__module__,
                obj.__qualname__,
            ],
        )
        return hash



from natch.decorators.any import any
from natch.decorators.any_of import any_of
from natch.decorators.all_of import all_of
from natch.decorators.eq import eq
from natch.decorators.neq import neq
from natch.decorators.gt import gt
from natch.decorators.gte import gte
from natch.decorators.lt import lt
from natch.decorators.lte import lte
from natch.decorators.contains import contains
from natch.decorators.not_contains import not_contains
from natch.decorators.condition import condition
from natch.decorators.partial import partial
from natch.decorators.pattern import pattern


__all__ = [
    'any',
    'any_of',
    'all_of',
    'eq',
    'neq',
    'gt',
    'gte',
    'lt',
    'lte',
    'contains',
    'not_contains',
    'condition',
    'partial',
    'pattern',
]



from natch.abstract import Rule
from natch.rules.eq import Eq


class Pattern(Rule):

    def __init__(self, *rules, **kwargs):
        super(Pattern, self).__init__(*rules)

    def set_args(self, args):
        self.del_args()
        for arg in args:
            if not isinstance(arg, Rule):
                arg = Eq(arg)
            self.args.append(arg)

    def does_match(self, *args, **kwargs):
        len_args = len(args)
        check_kwargs = bool(len(kwargs))
        if check_kwargs:
            len_args += 1
        if len(self.args) != len_args:
            return False
        for rule_idx, arg in enumerate(args):
            rule = self.args[rule_idx]
            has_match = rule.does_match(arg)
            if not has_match:
                return False
        if check_kwargs:
            rule = self.args[len_args - 1]
            has_match = rule.does_match(kwargs)
            if not has_match:
                return False
        return True



from natch.abstract import Rule
from natch.core.decoration import Decoration
from natch.exceptions import NeverMatchesError
from natch.decorators import any
from natch.decorators import any_of
from natch.decorators import all_of
from natch.decorators import eq
from natch.decorators import neq
from natch.decorators import gt
from natch.decorators import gte
from natch.decorators import lt
from natch.decorators import lte
from natch.decorators import contains
from natch.decorators import not_contains
from natch.decorators import condition
from natch.decorators import partial
from natch.decorators import pattern
from natch.rules import Any
from natch.rules import AnyOf
from natch.rules import AllOf
from natch.rules import Eq
from natch.rules import Neq
from natch.rules import Gt
from natch.rules import Gte
from natch.rules import Lt
from natch.rules import Lte
from natch.rules import Contains
from natch.rules import NotContains
from natch.rules import Condition
from natch.rules import Partial
from natch.rules import Pattern


__version__ = '1.0.4'


__all__ = [
    'any',
    'any_of',
    'all_of',
    'eq',
    'neq',
    'gt',
    'gte',
    'lt',
    'lte',
    'contains',
    'not_contains',
    'condition',
    'partial',
    'pattern',
    'Any',
    'AnyOf',
    'AllOf',
    'Eq',
    'Neq',
    'Gt',
    'Gte',
    'Lt',
    'Lte',
    'Contains',
    'NotContains',
    'Condition',
    'Partial',
    'Pattern',
]


make_rule_decorator = Decoration.make_rule_decorator



import abc


class Hasher(abc.ABC):

    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def hash(self, obj):
        raise NotImplementedError(
            f'{self.__class__.__name__}.hash method is not implemented.',
        )



from natch.abstract import Rule
from natch.rules.eq import Eq


class Partial(Rule):

    def __init__(self, *args, **kwargs):
        super(Partial, self).__init__(*args, **kwargs)

    def set_kwargs(self, kwargs):
        self.del_kwargs()
        for key, value in kwargs.items():
            if not isinstance(value, Rule):
                value = Eq(value)
            self.kwargs[key] = value

    def _does_match_object(self, obj):
        for key, rule in self.kwargs.items():
            has_attr = hasattr(obj, key)
            if not has_attr:
                return False
            partial_value = getattr(obj, key, None)
            does_match = rule.does_match(partial_value)
            if not does_match:
                return False
        return True

    def _does_match_dict(self, d):
        for key, rule in self.kwargs.items():
            if key not in d:
                return False
            partial_value = d.get(key)
            does_match = rule.does_match(partial_value)
            if not does_match:
                return False
        return True

    def does_match(self, *args, **kwargs):
        value = args[0] if len(args) else kwargs
        check = self._does_match_object
        if isinstance(value, dict):
            check = self._does_match_dict
        does_match = check(value)
        return does_match



from natch.abstract import Rule


class Gt(Rule):

    def __init__(self, value, **kwargs):
        super(Gt, self).__init__(*[value])

    def does_match(self, *args, **kwargs):
        value = args[0]
        return value > self.args[0]



import abc


class Rule(abc.ABC):

    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def get_args(self):
        return self._args

    def set_args(self, args):
        if args is None:
            args = list()
        self._args = list()
        for arg in args:
            self.args.append(arg)

    def del_args(self):
        self._args = list()

    @property
    def args(self):
        args = self.get_args()
        return args

    @args.setter
    def args(self, args):
        self.set_args(args)

    @args.deleter
    def args(self):
        self.del_args()

    def get_kwargs(self):
        return self._kwargs

    def set_kwargs(self, kwargs):
        if kwargs is None:
            kwargs = dict()
        self._kwargs = kwargs

    def del_kwargs(self):
        self._kwargs = dict()

    @property
    def kwargs(self):
        kwargs = self.get_kwargs()
        return kwargs

    @kwargs.setter
    def kwargs(self, kwargs):
        self.set_kwargs(kwargs)

    @kwargs.deleter
    def kwargs(self):
        self.del_kwargs()

    @abc.abstractmethod
    def does_match(self, *args, **kwargs):
        raise NotImplementedError(
            f'{self.__class__.__name__}.has_match method is not implemented.',
        )



from natch.core.globals import registry
from natch.exceptions import NeverMatchesError


class Decoration(object):

    @classmethod
    def make_rule_decorator(cls, rule_cls):
        def rule_decorator(*rule_args, **rule_kwargs):
            rule = rule_cls(*rule_args, **rule_kwargs)
            def func_wrapper(func):
                registry.register(func, rule)
                def executor(*args, **kwargs):
                    f = registry.lookup(func, *args, **kwargs)
                    if f is None:
                        raise NeverMatchesError()
                    result = f(*args, **kwargs)
                    return result
                return executor
            return func_wrapper
        return rule_decorator



from natch.abstract import Rule
from natch.rules.eq import Eq


class AllOf(Rule):

    def __init__(self, *args, **kwargs):
        super(AllOf, self).__init__(*args)

    def set_args(self, args):
        self.del_args()
        for arg in args:
            if not isinstance(arg, Rule):
                arg = Eq(arg)
            self.args.append(arg)

    def does_match(self, *args, **kwargs):
        value = args[0]
        for rule in self.args:
            does_match = rule.does_match(value)
            if not does_match:
                return False
        return True



from natch.abstract import Registry as AbstractRegistry
from natch.hashers import QualnameHasher


class Registry(AbstractRegistry):

    def __init__(self, *args, **kwargs):
        super(Registry, self).__init__(*args, **kwargs)

    def set_hasher(self, hasher):
        if hasher is None:
            hasher = QualnameHasher()
        super(Registry, self).set_hasher(hasher)

    def register(self, func, rule):
        func_hash = self.hasher.hash(func)
        if func_hash not in self.index:
            self.index[func_hash] = []
        path = (func, rule)
        self.index[func_hash].append(path)

    def unregister(self, func, rule):
        func_hash = self.hasher.hash(func)
        if func_hash not in self.index:
            return
        alternation = None
        for (f, r) in self.index[func_hash]:
            if r == rule:
                alternation = (f, r)
                break
        if alternation is not None:
            self.index[func_hash].remove(alternation)

    def lookup(self, func, *args, **kwargs):
        func_hash = self.hasher.hash(func)
        if func_hash not in self.index:
            return None
        for (f, rule) in self.index[func_hash]:
            does_match = rule.does_match(*args, **kwargs)
            if does_match:
                return f
        return None



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from setuptools import setup
import natch


setup(
    name='natch',
    version=natch.__version__,
    keywords='functional paradigm pattern matching',
    description='Pattern matching library.',
    license='MIT',
    author='Ertuğrul Keremoğlu',
    author_email='ertugkeremoglu@gmail.com',
    url='https://github.com/ertgl/natch/',
    classifiers=[
        'Topic :: Utilities',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: MIT License',
    ],
    packages=[
        'natch',
        'natch.abstract',
        'natch.core',
        'natch.decorators',
        'natch.exceptions',
        'natch.hashers',
        'natch.rules',
    ],
    install_requires=[],
)



import abc


class Registry(abc.ABC):

    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        self.index = kwargs.get('index')
        self.hasher = kwargs.get('hasher')

    def get_index(self):
        return self._index

    def set_index(self, index):
        if index is None:
            index = dict()
        self._index = index

    def del_index(self):
        self._index = dict()

    @property
    def index(self):
        index = self.get_index()
        return index

    @index.setter
    def index(self, index):
        self.set_index(index)

    @index.deleter
    def index(self):
        self.del_index()

    def get_hasher(self):
        return self._hasher

    def set_hasher(self, hasher):
        self._hasher = hasher

    def del_hasher(self):
        self._hasher = None

    @property
    def hasher(self):
        hasher = self.get_hasher()
        return hasher

    @hasher.setter
    def hasher(self, hasher):
        self.set_hasher(hasher)

    @hasher.deleter
    def hasher(self):
        self.del_hasher()

    @abc.abstractmethod
    def register(self, func, rule):
        raise NotImplementedError(
            f'{self.__class__.__name__}.register method is not implemented.',
        )

    @abc.abstractmethod
    def unregister(self, func, rule):
        raise NotImplementedError(
            f'{self.__class__.__name__}.unregister method is not implemented.',
        )

    @abc.abstractmethod
    def lookup(self, func, *args, **kwargs):
        raise NotImplementedError(
            f'{self.__class__.__name__}.lookup method is not implemented.',
        )



from natch.abstract import Rule


class Condition(Rule):

    def __init__(self, func, **kwargs):
        if not callable(func):
            raise TypeError('func must be a callable')
        super(Condition, self).__init__(*[func])

    def does_match(self, *args, **kwargs):
        value = args[0]
        result = self.args[0](value)
        return result
